@@ -13,16 +13,17 @@
 package Net::STOMP::Client;
 use strict;
 use warnings;
-our $VERSION  = "1.1_2";
-our $REVISION = sprintf("%d.%02d", q$Revision: 1.90 $ =~ /(\d+)\.(\d+)/);
+our $VERSION  = "1.6_2";
+our $REVISION = sprintf("%d.%03d", q$Revision: 1.101 $ =~ /(\d+)\.(\d+)/);
 
 #
 # used modules
 #
 
 use IO::Socket::INET;
-use List::Util qw(max shuffle);
+use List::Util qw(max);
 use Time::HiRes qw();
+use Net::STOMP::Client::Connection;
 use Net::STOMP::Client::Debug;
 use Net::STOMP::Client::Error;
 use Net::STOMP::Client::Frame;
@@ -37,7 +38,7 @@ use Net::STOMP::Client::Protocol qw(FLAG_DIRECTION_C2S FLAG_DIRECTION_S2C);
 use Net::STOMP::Client::OO;
 our(@ISA) = qw(Net::STOMP::Client::OO);
 Net::STOMP::Client::OO::methods(qw(
-    uri host port timeout sockopts callbacks peer session
+    uri host port timeout sockopts callbacks peer server session
     client_heart_beat server_heart_beat
     _io _id _serial _receipts _version
 ));
@@ -71,137 +72,31 @@ sub _timeout : method {
 	} else {
 	    # scalar timeout specified -> backward compatibility
 	    $timeout = {
-		connect   => $timeout,
-		connected => $timeout,
-		send      => undef,
-		receive   => undef,
+		connect    => $timeout,
+		connected  => $timeout,
+		disconnect => $timeout,
+		send       => undef,
+		receive    => undef,
 	    };
 	}
     } else {
 	# no timeout specified -> use hard-coded defaults
 	$timeout = {
-	    connect   => undef,
-	    connected => 10,
-	    send      => undef,
-	    receive   => undef,
+	    connect    => undef,
+	    connected  => 10,
+	    disconnect => 10,
+	    send       => undef,
+	    receive    => undef,
 	};
     }
-    unless ($operation =~ /^(connect|connected|receive|send)$/) {
-	Net::STOMP::Client::Error::report("%s: unexpected timeout operation: %s",
-					  "Net::STOMP::Client", $operation);
+    unless ($operation =~ /^(connect|connected|disconnect|receive|send)$/) {
+	Net::STOMP::Client::Error::report
+	    ("%s: unexpected timeout operation: %s", "Net::STOMP::Client", $operation);
 	return();
     }
     return($timeout->{$operation});
 }
 
-#
-# try to connect to the server (socket level only)
-#
-
-sub _new_socket ($$$$%) {
-    my($me, $proto, $host, $port, %sockopts) = @_;
-    my($socket);
-
-    # initial checks
-    $sockopts{Proto} = "tcp"; # yes, even SSL is TCP...
-    $sockopts{PeerAddr} = $host;
-    $sockopts{PeerPort} = $port;
-    # from now on, the errors are non-fatal and checked by the caller...
-    local $Net::STOMP::Client::Error::Die = 0;
-    # try to connect
-    if ($proto =~ /\b(ssl)\b/) {
-	# with SSL
-	unless ($INC{"IO/Socket/SSL.pm"}) {
-	    eval { require IO::Socket::SSL };
-	    if ($@) {
-		Net::STOMP::Client::Error::report("%s: cannot load IO::Socket::SSL: %s",
-						  $me, $@);
-		return();
-	    }
-	}
-	$socket = IO::Socket::SSL->new(%sockopts);
-	unless ($socket) {
-	    Net::STOMP::Client::Error::report("%s: cannot SSL connect to %s:%d: %s",
-				      $me, $sockopts{PeerAddr}, $sockopts{PeerPort},
-				      IO::Socket::SSL::errstr());
-	    return();
-	}
-    } else {
-	# with TCP
-	$socket = IO::Socket::INET->new(%sockopts);
-	unless ($socket) {
-	    Net::STOMP::Client::Error::report("%s: cannot connect to %s:%d: %s",
-				      $me, $sockopts{PeerAddr}, $sockopts{PeerPort}, $!);
-	    return();
-	}
-	unless (binmode($socket)) {
-	    Net::STOMP::Client::Error::report("%s: cannot binmode(socket): %s", $me, $!);
-	    return();
-	}
-    }
-    # so far so good...
-    return($socket);
-}
-
-#
-# check and handle the uri option
-#
-# see: http://activemq.apache.org/failover-transport-reference.html
-#
-
-sub _handle_uri ($$) {
-    my($me, $string) = @_;
-    my(@uris, $uri, $path, $fh, $line, @list, @peers);
-
-    @uris = ($string);
-    while (@uris) {
-	$uri = shift(@uris);
-	if ($uri =~ /^file:(.+)$/) {
-	    # list of uris stored in a file, one per line
-	    @list = ();
-	    $path = $1;
-	    unless (open($fh, "<", $path)) {
-		Net::STOMP::Client::Error::report("%s: cannot open %s: %s", $me, $path, $!);
-		return();
-	    }
-	    while (defined($line = <$fh>)) {
-		$line =~ s/\#.*//;
-		$line =~ s/\s+//g;
-		push(@list, $line) if length($line);
-	    }
-	    unless (close($fh)) {
-		Net::STOMP::Client::Error::report("%s: cannot close %s: %s", $me, $path, $!);
-		return();
-	    }
-	    unshift(@uris, @list);
-	} elsif ($uri =~ /^failover:(\/\/)?\(([\w\.\-\:\/\,]+)\)(\?[\w\=\-\&]+)?$/) {
-	    # failover with options (only "randomize" is supported)
-	    @list = split(/,/, $2);
-	    @list = shuffle(@list) unless $3 and $3 =~ /\b(randomize=false)\b/;
-	    unshift(@uris, @list);
-	} elsif ($uri =~ /^failover:([\w\.\-\:\/\,]+)$/) {
-	    # failover without options (randomized by default)
-	    @list = split(/,/, $1);
-	    unshift(@uris, shuffle(@list));
-	} elsif ($uri =~ /^(tcp|ssl|stomp|stomp\+ssl):\/\/([\w\.\-]+):(\d+)\/?$/) {
-	    # well-formed uri
-	    push(@peers, Net::STOMP::Client::Peer->new(
-		proto => $1,
-		host  => $2,
-		port  => $3,
-	    ));
-	} else {
-	    Net::STOMP::Client::Error::report("%s: unexpected server uri: %s", $me, $string);
-	    return();
-	}
-    }
-    unless (@peers) {
-	Net::STOMP::Client::Error::report("%s: empty server uri: %s", $me, $string);
-	return();
-    }
-    return(@peers);
-}
-
 #
 # check the version option
 #
@@ -226,8 +121,8 @@ sub _check_version ($$) {
     }
     foreach $version (@list) {
 	unless ($version =~ /^1\.[01]$/) {
-	    Net::STOMP::Client::Error::report("%s: unsupported STOMP version: %s",
-					      $me, $version);
+	    Net::STOMP::Client::Error::report
+		("%s: unsupported STOMP version: %s", $me, $version);
 	    return();
 	}
     }
@@ -242,7 +137,7 @@ sub _check_version ($$) {
 
 sub new : method {
     my($class, %data) = @_;
-    my($me, $self, %sockopts, $timeout, @uris, $uri, @peers, $peer, $socket, $io, $version);
+    my($me, $self, %sockopts, $timeout, $peer, $socket, $io, $version);
 
     $me = "Net::STOMP::Client->new()";
     $version = delete($data{version});
@@ -255,52 +150,10 @@ sub new : method {
 	$timeout = $self->_timeout("connect");
 	$sockopts{Timeout} = $timeout if $timeout;
     }
-    if ($self->uri()) {
-	if ($self->host()) {
-	    Net::STOMP::Client::Error::report("%s: unexpected server host: %s",
-					      $me, $self->host());
-	    return();
-	}
-	if ($self->port()) {
-	    Net::STOMP::Client::Error::report("%s: unexpected server port: %s",
-					      $me, $self->port());
-	    return();
-	}
-	@peers = _handle_uri($me, $self->uri());
-	return() unless @peers;
-	# use by default a shorter timeout in case we use failover...
-	if (@peers > 1 and not $sockopts{Timeout}) {
-	    # it makes no sense to block with failover!
-	    $sockopts{Timeout} = 3;
-	}
-    } else {
-	unless ($self->host()) {
-	    Net::STOMP::Client::Error::report("%s: missing server host", $me);
-	    return();
-	}
-	unless ($self->port()) {
-	    Net::STOMP::Client::Error::report("%s: missing server port", $me);
-	    return();
-	}
-	@peers = (Net::STOMP::Client::Peer->new(
-	    proto => grep(/^SSL_/, keys(%sockopts)) ? "ssl" : "tcp",
-	    host  => $self->host(),
-	    port  => $self->port(),
-	));
-    }
-    foreach $peer (@peers) {
-	$socket = _new_socket($me, $peer->proto(), $peer->host(), $peer->port(), %sockopts);
-	if ($socket) {
-	    # keep track of the peer we are connected to
-	    $peer->addr($socket->peerhost()); # real peer address
-	    $self->peer($peer);
-	    last;
-	}
-    }
-    unless ($socket) {
-	Net::STOMP::Client::Error::report("%s", $Net::STOMP::Client::Error::Message);
-	return();
-    }
+    ($socket, $peer) = Net::STOMP::Client::Connection::_new
+	($me, $self->host(), $self->port(), $self->uri(), %sockopts);
+    return() unless $socket;
+    $self->peer($peer);
     # final setup steps
     if ($self->uri()) {
 	# keep track of the peer this way too...
@@ -318,6 +171,7 @@ sub new : method {
     $self->_serial(0);
     $self->_receipts({});
     $self->callbacks({});
+    $self->server("");
     $self->session("");
     return($self);
 }
@@ -381,9 +235,11 @@ sub send_data : method {
     my($self, $timeout) = @_;
 
     # check that the I/O object is still usable
-    Net::STOMP::Client::Error::report(
-	"Net::STOMP::Client->send_data(): lost connection"
-    ) unless $self->_io();
+    unless ($self->_io()) {
+	Net::STOMP::Client::Error::report
+	    ("Net::STOMP::Client->send_data(): lost connection");
+	return();
+    }
     # handle the global send timeout
     $timeout = $self->_timeout("send") unless defined($timeout);
     # just do it
@@ -406,9 +262,8 @@ sub send_frame : method {
     return() unless defined($result);
     # make sure we sent _all_ data
     if ($self->_io()->outgoing_buffer_length()) {
-	Net::STOMP::Client::Error::report(
-	    "Net::STOMP::Client->send_frame(): could not send all data!"
-	);
+	Net::STOMP::Client::Error::report
+	    ("Net::STOMP::Client->send_frame(): could not send all data!");
 	return();
     }
     # so far so good...
@@ -423,9 +278,11 @@ sub receive_data : method {
     my($self, $timeout) = @_;
 
     # check that the I/O object is still usable
-    Net::STOMP::Client::Error::report(
-	"Net::STOMP::Client->receive_data(): lost connection"
-    ) unless $self->_io();
+    unless ($self->_io()) {
+	Net::STOMP::Client::Error::report
+	    ("Net::STOMP::Client->receive_data(): lost connection");
+	return();
+    }
     # handle the global receive timeout
     $timeout = $self->_timeout("receive") unless defined($timeout);
     # just do it
@@ -666,8 +523,8 @@ sub _unexpected_frame ($$) {
     my($frame, $info) = @_;
 
     $info ||= "?";
-    Net::STOMP::Client::Error::report("unexpected %s frame received: %s",
-				      $frame->command(), $info);
+    Net::STOMP::Client::Error::report
+	("unexpected %s frame received: %s", $frame->command(), $info);
 }
 
 #
@@ -679,6 +536,11 @@ sub _default_connected_callback ($$) {
     my($self, $frame) = @_;
     my($value);
 
+    # report an error if we are already connected
+    if ($self->session()) {
+	_unexpected_frame($frame, "already connected");
+	return();
+    }
     # STOMP 1.1: handle version negotiation
     $value = $frame->header("version");
     if ($value) {
@@ -724,16 +586,15 @@ sub _default_connected_callback ($$) {
 	$self->client_heart_beat(0);
 	$self->server_heart_beat(0);
     }
+    # record the server header
+    $value = $frame->header("server");
+    $self->server($value) if $value;
     # keep track of session id
-    unless ($self->session()) {
-	$value = $frame->header("session");
-	# the session header is optional so we forge our own if it is missing
-	$value ||= sprintf("sid-%s", $self->_id());
-	$self->session($value);
-	return($self);
-    }
-    _unexpected_frame($frame, $frame->header("session"));
-    return();
+    $value = $frame->header("session");
+    # the session header is optional so we forge our own if it is missing
+    $value ||= sprintf("sid-%s", $self->_id());
+    $self->session($value);
+    return($self);
 }
 $Callback{CONNECTED} = \&_default_connected_callback;
 
@@ -852,8 +713,8 @@ sub connect : method {
     }
     # must be called before connection
     if ($self->session()) {
-	Net::STOMP::Client::Error::report("%s: already connected with session: %s",
-					  $me, $self->session());
+	Net::STOMP::Client::Error::report
+	    ("%s: already connected with session: %s", $me, $self->session());
 	return();
     }
     # so far so good
@@ -884,8 +745,8 @@ sub connect : method {
     );
     return() unless defined($session);
     return($self) if $session;
-    Net::STOMP::Client::Error::report("Net::STOMP::Client->connect(): %s",
-				      "no CONNECTED frame received");
+    Net::STOMP::Client::Error::report
+	("Net::STOMP::Client->connect(): %s", "no CONNECTED frame received");
     return();
 }
 
@@ -905,6 +766,13 @@ sub disconnect : method {
 	headers => \%option,
     );
     $self->send_frame($frame, $timeout) or return();
+    # if a receipt has been given, wait for it!
+    if ($option{receipt}) {
+	$self->wait_for_frames(
+	    timeout  => $self->_timeout("disconnect"),
+	    callback => sub { return(! $self->_receipts()->{$option{receipt}}) },
+	);
+    }
     # additional bookkeeping
     $self->session("");
     $self->peer(undef);
@@ -1010,8 +878,8 @@ sub nack : method {
 
     $self->_check_invocation(scalar(@_)) or return();
     if ($self->version() eq "1.0") {
-	Net::STOMP::Client::Error::report("Net::STOMP::Client->nack(): %s %s",
-					  "not supported for STOMP", $self->version());
+	Net::STOMP::Client::Error::report
+	    ("Net::STOMP::Client->nack(): %s %s", "not supported for STOMP", $self->version());
 	return();
     }
     $timeout = delete($option{timeout});
@@ -1163,7 +1031,11 @@ Net::STOMP::Client - STOMP object oriented client module
       return($self);
   });
   # subscribe to the given queue
-  $stomp->subscribe(destination => "/queue/test", ack => "client");
+  $stomp->subscribe(
+      destination => "/queue/test",
+      id          => "testsub",          # required in STOMP 1.1
+      ack         => "client",           # client side acknowledgment
+  );
   # wait for a specified message frame
   $stomp->wait_for_frames(callback => sub {
       my($self, $frame) = @_;
@@ -1175,7 +1047,7 @@ Net::STOMP::Client - STOMP object oriented client module
       # continue to wait for more frames
       return(0);
   });
-  $stomp->unsubscribe(destination => "/queue/test");
+  $stomp->unsubscribe(id => "testsub");
   $stomp->disconnect();
 
 =head1 DESCRIPTION
@@ -1202,8 +1074,8 @@ list of strings); this defaults to the list of all supported versions
 
 the Uniform Resource Identifier (URI) specifying where the STOMP
 service is and how to connect to it, this can be for instance
-C<tcp://msg01:6163> or something more complex such as
-C<failover:(ssl://msg01:6162,tcp://msg01:6163)>
+C<tcp://msg01:6163> or something more complex, see the L</"FAILOVER">
+section for more information
 
 =item C<host>
 
@@ -1242,7 +1114,26 @@ section for more information
 Upon object creation, a TCP connection is made to the server but no
 data (i.e. STOMP frame) is exchanged.
 
-=head1 SSL
+=head2 FAILOVER
+
+The C<uri> option of the new() method can be given a complex URI
+indicating some kind of failover, for instance:
+C<failover:(tcp://msg01:6163,tcp://msg02:6163)>.
+
+This must use the ActiveMQ syntax
+(see L<http://activemq.apache.org/failover-transport-reference.html>)
+and only some options are supported, namely: C<backOffMultiplier>,
+C<initialReconnectDelay>, C<maxReconnectAttempts>,
+C<maxReconnectDelay>, C<randomize> and C<useExponentialBackOff>.
+
+When specified, these failover options will be used only inside the
+new() method (so at the TCP connection level) and not elsewhere. If
+the broker later fails during the STOMP interaction, it is up to the
+program author, knowing the logic of his code, to perform the
+appropriate recovery actions and eventually reconnect, using again the
+new() method.
+
+=head2 SSL
 
 When creating an object with Net::STOMP::Client->new(), if you supply some
 socket options (via C<sockopts>) with a name starting with C<SSL_>,
@@ -1277,7 +1168,7 @@ private key
 
 For more information, see L<IO::Socket::SSL>.
 
-=head1 TIMEOUTS
+=head2 TIMEOUTS
 
 By default, when sending STOMP frames, the module waits until the
 frame indeed has been sent (from the socket point of view). In case
@@ -1301,6 +1192,12 @@ L<IO::Socket::INET> or L<IO::Socket::SSL> object (default: none)
 timeout used while waiting for the initial CONNECTED frame from the
 broker (default: 10)
 
+=item disconnect
+
+timeout specifying how long the disconnect() method should wait for a
+RECEIPT frame back in case the DISCONNECT frame contained a receipt
+(default: 10)
+
 =item receive
 
 timeout used while trying to receive any frame (default: none)
@@ -1373,6 +1270,7 @@ frame headers. For instance:
 
   $stomp->subscribe(
       destination => "/queue/test",
+      id          => "testsub",
       ack         => "client",
   );
 
@@ -1417,6 +1315,11 @@ the connected STOMP server
 
 return the file handle of the socket connecting the client and the server
 
+=item server()
+
+return the server header seen on the CONNECTED frame or the empty
+string if not seen
+
 =item session()
 
 return the session identifier if connected or the empty string otherwise
@@ -1681,14 +1584,14 @@ positive number (meaning block at most this number of seconds).
 
 =head1 COMPATIBILITY
 
-This module implements the versions C<1.0> (see
-L<http://stomp.github.com/stomp-specification-1.0.html>) and C<1.1>
-(see L<http://stomp.github.com/stomp-specification-1.1.html>) of the
-protocol as well as well known extensions for JMS, AMQP, ActiveMQ and
-RabbitMQ.
+This module implements the versions
+C<1.0> (see L<http://stomp.github.com/stomp-specification-1.0.html>) and
+C<1.1> (see L<http://stomp.github.com/stomp-specification-1.1.html>)
+of the protocol as well as well known extensions for JMS, ActiveMQ,
+Apollo and RabbitMQ.
 
-It has been successfully tested against both ActiveMQ and RabbitMQ
-brokers.
+It has been successfully tested against ActiveMQ, Apollo, HornetQ and
+RabbitMQ brokers.
 
 =head1 SEE ALSO
 
@@ -1703,4 +1606,4 @@ L<Time::HiRes>.
 
 Lionel Cons L<http://cern.ch/lionel.cons>
 
-Copyright CERN 2010-2011
+Copyright CERN 2010-2012
