@@ -13,8 +13,8 @@
 package Net::STOMP::Client::Frame;
 use strict;
 use warnings;
-our $VERSION  = "1.1_2";
-our $REVISION = sprintf("%d.%02d", q$Revision: 1.49 $ =~ /(\d+)\.(\d+)/);
+our $VERSION  = "1.6_2";
+our $REVISION = sprintf("%d.%02d", q$Revision: 1.54 $ =~ /(\d+)\.(\d+)/);
 
 #
 # Object Oriented definition
@@ -273,6 +273,7 @@ sub decode ($%) {
     if (length($temp)) {
 	# optionally handle UTF-8 header decoding
 	if ($UTF8Header or (not defined($UTF8Header) and $v1_1)) {
+	    local $@; # preserve $@!
 	    $temp = Encode::decode("UTF-8", $temp, $fb);
 	}
 	if ($v1_1) {
@@ -317,6 +318,7 @@ sub decode ($%) {
     # optionally handle body decoding
     if ($BodyEncode or (not defined($BodyEncode) and $v1_1)) {
 	$value = _encoding($headers->{"content-type"});
+	local $@; # preserve $@!
 	$temp = Encode::decode($value, $temp, $fb) if $value;
     }
     $frame->body_reference(\$temp);
@@ -347,6 +349,7 @@ sub encode : method {
     # optionally handle body encoding (before content-length handling)
     if ($BodyEncode or (not defined($BodyEncode) and $v1_1)) {
 	$string = _encoding($headers->{"content-type"});
+	local $@; # preserve $@!
 	$body = Encode::encode($string, $body, $fb) if $string;
     }
 
@@ -362,23 +365,27 @@ sub encode : method {
 	    unless $body eq "";
     }
 
-    # encode the command and header
-    $string = "";
-    while (($key, $value) = each(%$headers)) {
-	next if $key eq "content-length";
-	if ($v1_1) {
-	    # handle backslash escaping
-	    $key   =~ s/([\x0a\x3a\x5c])/$_EncMap{$1}/eg;
-	    $value =~ s/([\x0a\x3a\x5c])/$_EncMap{$1}/eg;
-	}
-	$string .= $key . ":" . $value . "\n";
+    # encode the header
+    if ($v1_1) {
+	# with backslash escaping
+	$string = join("\n", map({
+	    ($key = $_)               =~ s/([\x0a\x3a\x5c])/$_EncMap{$1}/eg;
+	    ($value = $headers->{$_}) =~ s/([\x0a\x3a\x5c])/$_EncMap{$1}/eg;
+	    "${key}:${value}";
+	} grep($_ ne "content-length", keys(%$headers))), "");
+    } else {
+	# without backslash escaping
+	$string = join("\n", map({
+	    "${_}:$headers->{$_}";
+	} grep($_ ne "content-length", keys(%$headers))), "");
     }
     if (defined($content_length)) {
-	$string .= "content-length:" . $content_length . "\n";
+	$string .= "content-length:${content_length}\n";
     }
 
     # optionally handle UTF-8 header encoding
     if ($UTF8Header or (not defined($UTF8Header) and $v1_1)) {
+	local $@; # preserve $@!
 	$string = Encode::encode("UTF-8", $string, $fb);
     }
 
@@ -393,7 +400,7 @@ sub encode : method {
     }
 
     # assemble all the parts
-    $value = $self->command() . "\n" . $string . "\n" . $body . "\0";
+    $value = $self->command() . "\n${string}\n${body}\0";
 
     # return a reference to the encoded frame
     return(\$value);
@@ -423,7 +430,7 @@ sub _debug_header ($) {
 	    if ($char == 0x0a) {
 		# end of header line
 		last;
-	    } elsif (0x20 <= $char and $char <= 0x7e and $char != 0x25) {
+	    } elsif (0x20 < $char and $char <= 0x7e and $char != 0x25) {
 		# printable
 		$line .= sprintf("%c", $char);
 	    } else {
@@ -628,6 +635,93 @@ sub check : method {
     return($self);
 }
 
+#+++############################################################################
+#                                                                              #
+# integration with Messaging::Message                                          #
+#                                                                              #
+#---############################################################################
+
+#
+# transform a frame into a message
+#
+
+sub messagify : method {
+    my($self) = @_;
+    my($me, $msg, $tmp);
+
+    $me = "Net::STOMP::Client::Frame->messagify()";
+    unless ($Messaging::Message::VERSION) {
+	eval { require Messaging::Message };
+	if ($@) {
+	    Net::STOMP::Client::Error::report("%s: cannot load Messaging::Message: %s",
+					      $me, $@);
+	    return();
+	}
+    }
+    $msg = Messaging::Message->new();
+    # handler header
+    $tmp = $self->headers();
+    $msg->header($tmp) if $tmp;
+    # handle text
+    $tmp = $msg->header_field("content-type");
+    $msg->text(1) if $tmp and ($tmp =~ /^text\// or $tmp =~ /\bcharset=/);
+    # handle body
+    $tmp = $self->body_reference();
+    $msg->body_ref($tmp) if $tmp;
+    return($msg);
+}
+
+#
+# transform a message into a frame
+#
+
+sub demessagify : method {
+    my($class, $msg) = @_;
+    my($me, $frame, $tmp);
+
+    $me = "Net::STOMP::Client::Frame->demessagify()";
+    unless ($Messaging::Message::VERSION) {
+	eval { require Messaging::Message };
+	if ($@) {
+	    Net::STOMP::Client::Error::report("%s: cannot load Messaging::Message: %s",
+					      $me, $@);
+	    return();
+	}
+    }
+    unless ($msg and ref($msg) and $msg->isa("Messaging::Message")) {
+	Net::STOMP::Client::Error::report("%s: missing message", $me);
+	return();
+    }
+    $frame = Net::STOMP::Client::Frame->new(
+	command        => "SEND",
+	headers        => $msg->header(),
+	body_reference => $msg->body_ref(),
+    );
+    # handle text
+    $tmp = $msg->header_field("content-type");
+    if (defined($tmp)) {
+	# make sure the content-type is consistent with the message type
+	if ($tmp =~ /^text\// or $tmp =~ /\bcharset=/) {
+	    unless ($msg->text()) {
+		Net::STOMP::Client::Error::report
+		    ("%s: unexpected text content-type for binary message: %s", $me, $tmp);
+		return();
+	    }
+	} else {
+	    if ($msg->text()) {
+		Net::STOMP::Client::Error::report
+		    ("%s: unexpected binary content-type for text message: %s", $me, $tmp);
+		return();
+	    }
+	}
+    } else {
+	# set a text content-type if it is missing (this is needed by STOMP 1.1)
+	$frame->headers({ %{ $frame->headers() }, "content-type" => "text/unknown" })
+	    if $msg->text();
+    }
+    return($frame);
+}
+
 1;
 
 __END__
@@ -846,6 +940,37 @@ Encode::FB_CROAK flag instead of the default Encode::FB_DEFAULT.
 N.B.: Perl's standard Encode module is used for all encoding/decoding
 operations.
 
+=head1 MESSAGING ABSTRACTION
+
+If the L<Messaging::Message> module is available, the following
+methods are available:
+
+=over
+
+=item messagify()
+
+transform the frame into a Messaging::Message object
+
+=item demessagify(MESSAGE)
+
+transform the given Messaging::Message object into a
+Net::STOMP::Client::Frame object (class method)
+
+=back
+
+Here is how they could be used:
+
+  # frame to message
+  $frame = $stomp->wait_for_frames(timeout => 1);
+  if ($frame) {
+      $message = $frame->messagify();
+      ...
+  }
+
+  # message to frame
+  $frame = Net::STOMP::Client::Frame->demessagify($message);
+  $stomp->send_frame($frame);
+
 =head1 COMPLIANCE
 
 STOMP 1.0 has several ambiguities and this module does its best to
@@ -961,10 +1086,11 @@ method.
 
 L<Net::STOMP::Client::Debug>,
 L<Net::STOMP::Client::OO>,
+L<Messaging::Message>,
 L<Encode>.
 
 =head1 AUTHOR
 
 Lionel Cons L<http://cern.ch/lionel.cons>
 
-Copyright CERN 2010-2011
+Copyright CERN 2010-2012
